@@ -24,9 +24,14 @@
 #endif
 
 #include <stdio.h>
+#include <unistd.h>
 #include <errno.h>
 #include <stdlib.h>
 #include <string.h>
+#include <sys/ioctl.h>
+#include <sys/socket.h>
+#include <linux/if_arp.h>
+#include <linux/wireless.h>
 #include <net/ethernet.h>
 
 #include <gdbus.h>
@@ -172,12 +177,15 @@ struct supplicant_task {
 	char *ifname;
 	struct connman_device *device;
 	struct connman_network *network;
+	struct connman_network *pending_network;
 	char *path;
 	char *netpath;
 	gboolean created;
 	enum supplicant_state state;
 	gboolean noscan;
 	GSList *scan_results;
+	struct iw_range *range;
+	gboolean disconnecting;
 };
 
 static GSList *task_list = NULL;
@@ -221,6 +229,75 @@ static struct supplicant_task *find_task_by_path(const char *path)
 	return NULL;
 }
 
+static int get_range(struct supplicant_task *task)
+{
+	struct iwreq wrq;
+	int fd, err;
+
+	fd = socket(PF_INET, SOCK_DGRAM, 0);
+	if (fd < 0)
+		return -1;
+
+	memset(&wrq, 0, sizeof(struct iwreq));
+	strncpy(wrq.ifr_name, task->ifname, IFNAMSIZ);
+	wrq.u.data.pointer = task->range;
+	wrq.u.data.length = sizeof(struct iw_range);
+
+	err = ioctl(fd, SIOCGIWRANGE, &wrq);
+
+	close(fd);
+
+	return err;
+}
+
+static char *get_bssid(struct connman_device *device)
+{
+	char *bssid;
+	unsigned char ioctl_bssid[ETH_ALEN];
+	int ifindex;
+	char *ifname;
+	struct iwreq wrq;
+	int fd, err;
+
+	ifindex = connman_device_get_index(device);
+	if (ifindex < 0)
+		return NULL;
+
+	ifname = connman_inet_ifname(ifindex);
+	if (ifname == NULL)
+		return NULL;
+
+	fd = socket(PF_INET, SOCK_DGRAM, 0);
+	if (fd < 0) {
+		g_free(ifname);
+		return NULL;
+	}
+
+	memset(&wrq, 0, sizeof(wrq));
+	strncpy(wrq.ifr_name, ifname, IFNAMSIZ);
+
+	err = ioctl(fd, SIOCGIWAP, &wrq);
+
+	g_free(ifname);
+	close(fd);
+
+	if (err < 0)
+		return NULL;
+
+	memcpy(ioctl_bssid, wrq.u.ap_addr.sa_data, ETH_ALEN);
+
+	bssid = g_try_malloc0(13);
+	if (bssid == NULL)
+		return NULL;
+
+	snprintf(bssid, 13, "%02x%02x%02x%02x%02x%02x",
+					ioctl_bssid[0], ioctl_bssid[1],
+					ioctl_bssid[2], ioctl_bssid[3],
+					ioctl_bssid[4], ioctl_bssid[5]);
+
+	return bssid;
+}
+
 static void add_interface_reply(DBusPendingCall *call, void *user_data)
 {
 	struct supplicant_task *task = user_data;
@@ -707,6 +784,7 @@ static int set_network(struct supplicant_task *task,
 	DBusMessage *message, *reply;
 	DBusMessageIter array, dict;
 	DBusError error;
+	dbus_uint32_t scan_ssid = 1;
 
 	DBG("task %p", task);
 
@@ -725,11 +803,10 @@ static int set_network(struct supplicant_task *task,
 			DBUS_TYPE_STRING_AS_STRING DBUS_TYPE_VARIANT_AS_STRING
 			DBUS_DICT_ENTRY_END_CHAR_AS_STRING, &dict);
 
-	if (address == NULL) {
-		dbus_uint32_t scan_ssid = 1;
-		connman_dbus_dict_append_variant(&dict, "scan_ssid",
-						DBUS_TYPE_UINT32, &scan_ssid);
-	} else
+	connman_dbus_dict_append_variant(&dict, "scan_ssid",
+					 DBUS_TYPE_UINT32, &scan_ssid);
+
+	if (address)
 		connman_dbus_dict_append_variant(&dict, "bssid",
 						DBUS_TYPE_STRING, &address);
 
@@ -746,14 +823,16 @@ static int set_network(struct supplicant_task *task,
 			connman_dbus_dict_append_variant(&dict, "psk",
 						DBUS_TYPE_STRING, &passphrase);
 	} else if (g_ascii_strcasecmp(security, "wep") == 0) {
-		const char *key_mgmt = "NONE", *index = "0";
+		const char *key_mgmt = "NONE";
 		const char *auth_alg = "OPEN SHARED";
-		connman_dbus_dict_append_variant(&dict, "key_mgmt",
-						DBUS_TYPE_STRING, &key_mgmt);
+		const char *key_index = "0";
 
 		connman_dbus_dict_append_variant(&dict, "auth_alg",
 						DBUS_TYPE_STRING, &auth_alg);
 
+		connman_dbus_dict_append_variant(&dict, "key_mgmt",
+						DBUS_TYPE_STRING, &key_mgmt);
+
 		if (passphrase) {
 			int size = strlen(passphrase);
 			if (size == 10 || size == 26) {
@@ -776,8 +855,9 @@ static int set_network(struct supplicant_task *task,
 				connman_dbus_dict_append_variant(&dict,
 						"wep_key0", DBUS_TYPE_STRING,
 								&passphrase);
+
 			connman_dbus_dict_append_variant(&dict, "wep_tx_keyidx",
-						DBUS_TYPE_STRING, &index);
+						DBUS_TYPE_STRING, &key_index);
 		}
 	} else {
 		const char *key_mgmt = "NONE";
@@ -868,6 +948,11 @@ static char *build_group(const char *addr, const char *name,
 	if (str == NULL)
 		return NULL;
 
+	if (ssid == NULL) {
+		g_string_append_printf(str, "hidden_%s", addr);
+		goto done;
+	}
+
 	for (i = 0; special_ssid[i].name; i++) {
 		if (g_strcmp0(special_ssid[i].name, name) == 0) {
 			if (special_ssid[i].value == NULL)
@@ -896,7 +981,7 @@ static void extract_addr(DBusMessageIter *value,
 					struct supplicant_result *result)
 {
 	DBusMessageIter array;
-	struct ether_addr *eth;
+	struct ether_addr eth;
 	unsigned char *addr;
 	int addr_len;
 
@@ -917,15 +1002,14 @@ static void extract_addr(DBusMessageIter *value,
 	if (result->path == NULL)
 		return;
 
-	eth = (void *) addr;
-
+	memcpy(&eth, addr, sizeof(eth));
 	snprintf(result->path, 13, "%02x%02x%02x%02x%02x%02x",
-						eth->ether_addr_octet[0],
-						eth->ether_addr_octet[1],
-						eth->ether_addr_octet[2],
-						eth->ether_addr_octet[3],
-						eth->ether_addr_octet[4],
-						eth->ether_addr_octet[5]);
+						eth.ether_addr_octet[0],
+						eth.ether_addr_octet[1],
+						eth.ether_addr_octet[2],
+						eth.ether_addr_octet[3],
+						eth.ether_addr_octet[4],
+						eth.ether_addr_octet[5]);
 }
 
 static void extract_ssid(DBusMessageIter *value,
@@ -941,6 +1025,9 @@ static void extract_ssid(DBusMessageIter *value,
 	if (ssid_len < 1)
 		return;
 
+	if (ssid[0] == '\0')
+		return;
+
 	result->ssid = g_try_malloc(ssid_len);
 	if (result->ssid == NULL)
 		return;
@@ -1016,9 +1103,10 @@ static void extract_capabilites(DBusMessageIter *value,
 		result->has_wep = TRUE;
 }
 
-static unsigned char calculate_strength(struct supplicant_result *result)
+static unsigned char calculate_strength(struct supplicant_task *task,
+					struct supplicant_result *result)
 {
-	if (result->quality < 0) {
+	if (task->range->max_qual.qual == 0) {
 		unsigned char strength;
 
 		if (result->level > 0)
@@ -1032,7 +1120,7 @@ static unsigned char calculate_strength(struct supplicant_result *result)
 		return strength;
 	}
 
-	return result->quality;
+	return (result->quality * 100) / task->range->max_qual.qual;
 }
 
 static unsigned short calculate_channel(struct supplicant_result *result)
@@ -1055,7 +1143,8 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 	unsigned char strength;
 	unsigned short channel, frequency;
 	const char *mode, *security;
-	char *group;
+	char *group = NULL;
+	unsigned int ssid_len;
 
 	DBG("task %p", task);
 
@@ -1146,7 +1235,7 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 	else if (result.frequency == 14)
 		result.frequency = 2484;
 
-	strength = calculate_strength(&result);
+	strength = calculate_strength(task, &result);
 	channel  = calculate_channel(&result);
 
 	frequency = (result.frequency < 0) ? 0 : result.frequency;
@@ -1193,8 +1282,10 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 	if (result.name != NULL && result.name[0] != '\0')
 		connman_network_set_name(network, result.name);
 
-	connman_network_set_blob(network, "WiFi.SSID",
-						result.ssid, result.ssid_len);
+	if (connman_network_get_blob(network, "WiFi.SSID", &ssid_len) == NULL) {
+		connman_network_set_blob(network, "WiFi.SSID",
+					 result.ssid, result.ssid_len);
+	}
 
 	connman_network_set_string(network, "WiFi.Mode", mode);
 
@@ -1209,11 +1300,12 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 	connman_network_set_uint16(network, "WiFi.Channel", channel);
 	connman_network_set_string(network, "WiFi.Security", security);
 
-	connman_network_set_group(network, group);
-
-	g_free(group);
+	if (result.ssid != NULL)
+		connman_network_set_group(network, group);
 
 done:
+	g_free(group);
+
 	g_free(result.path);
 	g_free(result.addr);
 	g_free(result.name);
@@ -1264,8 +1356,7 @@ static void get_properties(struct supplicant_task *task)
 	return;
 
 noscan:
-	if (task->noscan == FALSE)
-		connman_device_set_scanning(task->device, FALSE);
+	connman_device_set_scanning(task->device, FALSE);
 }
 
 static void scan_results_reply(DBusPendingCall *call, void *user_data)
@@ -1322,8 +1413,7 @@ done:
 	dbus_message_unref(reply);
 
 noscan:
-	if (task->noscan == FALSE)
-		connman_device_set_scanning(task->device, FALSE);
+	connman_device_set_scanning(task->device, FALSE);
 }
 
 static void scan_results_available(struct supplicant_task *task)
@@ -1381,6 +1471,39 @@ static enum supplicant_state string2state(const char *state)
 		return WPA_INVALID;
 }
 
+static int task_connect(struct supplicant_task *task)
+{
+	const char *address, *security, *passphrase;
+	const void *ssid;
+	unsigned int ssid_len;
+
+	address = connman_network_get_string(task->network, "Address");
+	security = connman_network_get_string(task->network, "WiFi.Security");
+	passphrase = connman_network_get_string(task->network, "WiFi.Passphrase");
+
+	ssid = connman_network_get_blob(task->network, "WiFi.SSID", &ssid_len);
+
+	DBG("address %s security %s passphrase %s",
+					address, security, passphrase);
+
+	if (security == NULL && passphrase == NULL)
+		return -EINVAL;
+
+	if (g_str_equal(security, "none") == FALSE && passphrase == NULL)
+		return -EINVAL;
+
+	add_network(task);
+
+	select_network(task);
+	disable_network(task);
+
+	set_network(task, ssid, ssid_len, address, security, passphrase);
+
+	enable_network(task);
+
+	return -EINPROGRESS;
+}
+
 static void state_change(struct supplicant_task *task, DBusMessage *msg)
 {
 	DBusError error;
@@ -1436,18 +1559,76 @@ static void state_change(struct supplicant_task *task, DBusMessage *msg)
 
 	switch (task->state) {
 	case WPA_COMPLETED:
+		if (connman_network_get_group(task->network) == NULL) {
+			const char *name, *mode, *security;
+			char *bssid;
+
+			/*
+			 * This is a hidden network, we need to set its
+			 * group based on the BSSID we just joined.
+			 */
+			bssid = get_bssid(task->device);
+
+			name = connman_network_get_string(task->network,
+								"Name");
+			mode = connman_network_get_string(task->network,
+								"WiFi.Mode");
+			security = connman_network_get_string(task->network,
+							"WiFi.Security");
+
+			if (bssid && name && mode && security) {
+				char *group;
+
+				group = build_group(bssid, name, NULL, 0,
+								mode, security);
+				connman_network_set_group(task->network, group);
+				g_free(group);
+			}
+
+			g_free(bssid);
+		}
+
 		/* carrier on */
 		connman_network_set_connected(task->network, TRUE);
 		connman_device_set_scanning(task->device, FALSE);
 		break;
+
 	case WPA_DISCONNECTED:
-		/* carrier off */
-		connman_network_set_connected(task->network, FALSE);
-		connman_device_set_scanning(task->device, FALSE);
+		if (task->disconnecting == TRUE) {
+			connman_network_set_connected(task->network, FALSE);
+			connman_network_unref(task->network);
+			task->disconnecting = FALSE;
+
+			if (task->pending_network != NULL) {
+				task->network = task->pending_network;
+				task->pending_network = NULL;
+				task_connect(task);
+			}
+		} else {
+			/* carrier off */
+			connman_network_set_connected(task->network, FALSE);
+			connman_device_set_scanning(task->device, FALSE);
+		}
 		break;
+
 	case WPA_ASSOCIATING:
 		connman_network_set_associating(task->network, TRUE);
 		break;
+
+	case WPA_INACTIVE:
+		if (task->disconnecting == TRUE) {
+			connman_network_set_connected(task->network, FALSE);
+			connman_network_unref(task->network);
+			task->disconnecting = FALSE;
+
+			if (task->pending_network != NULL) {
+				task->network = task->pending_network;
+				task->pending_network = NULL;
+				task_connect(task);
+			}
+		}
+		break;
+
 	default:
 		connman_network_set_associating(task->network, FALSE);
 		break;
@@ -1489,6 +1670,7 @@ static DBusHandlerResult supplicant_filter(DBusConnection *conn,
 int supplicant_start(struct connman_device *device)
 {
 	struct supplicant_task *task;
+	int err;
 
 	DBG("device %p", device);
 
@@ -1500,19 +1682,38 @@ int supplicant_start(struct connman_device *device)
 	task->ifname = connman_inet_ifname(task->ifindex);
 
 	if (task->ifname == NULL) {
-		g_free(task);
-		return -ENOMEM;
+		err = -ENOMEM;
+		goto failed;
 	}
 
+	task->range = g_try_malloc0(sizeof(struct iw_range));
+	if (task->range == NULL) {
+		err = -ENOMEM;
+		goto failed;
+	}
+
+	err = get_range(task);
+	if (err < 0)
+		goto failed;
+
 	task->device = connman_device_ref(device);
 
 	task->created = FALSE;
 	task->noscan = FALSE;
 	task->state = WPA_INVALID;
+	task->disconnecting = FALSE;
+	task->pending_network = NULL;
 
 	task_list = g_slist_append(task_list, task);
 
 	return create_interface(task);
+
+failed:
+	g_free(task->range);
+	g_free(task->ifname);
+	g_free(task);
+
+	return err;
 }
 
 int supplicant_stop(struct connman_device *device)
@@ -1526,6 +1727,8 @@ int supplicant_stop(struct connman_device *device)
 	if (task == NULL)
 		return -ENODEV;
 
+	g_free(task->range);
+
 	task_list = g_slist_remove(task_list, task);
 
 	disable_network(task);
@@ -1567,48 +1770,24 @@ int supplicant_scan(struct connman_device *device)
 int supplicant_connect(struct connman_network *network)
 {
 	struct supplicant_task *task;
-	const char *address, *security, *passphrase;
-	const void *ssid;
-	unsigned int ssid_len;
 	int index;
 
 	DBG("network %p", network);
 
-	address = connman_network_get_string(network, "Address");
-	security = connman_network_get_string(network, "WiFi.Security");
-	passphrase = connman_network_get_string(network, "WiFi.Passphrase");
-
-	ssid = connman_network_get_blob(network, "WiFi.SSID", &ssid_len);
-
-	DBG("address %s security %s passphrase %s",
-					address, security, passphrase);
-
-	if (security == NULL && passphrase == NULL)
-		return -EINVAL;
-
-	if (g_str_equal(security, "none") == FALSE && passphrase == NULL)
-		return -EINVAL;
-
 	index = connman_network_get_index(network);
 
 	task = find_task_by_index(index);
 	if (task == NULL)
 		return -ENODEV;
 
-	task->network = connman_network_ref(network);
+	if (task->disconnecting == TRUE)
+		task->pending_network = connman_network_ref(network);
+	else {
+		task->network = connman_network_ref(network);
+		return task_connect(task);
+	}
 
-	add_network(task);
-
-	select_network(task);
-	disable_network(task);
-
-	set_network(task, ssid, ssid_len, address, security, passphrase);
-
-	enable_network(task);
-
-	connman_network_set_associating(task->network, TRUE);
-
-	return 0;
+	return -EINPROGRESS;
 }
 
 int supplicant_disconnect(struct connman_network *network)
@@ -1624,13 +1803,14 @@ int supplicant_disconnect(struct connman_network *network)
 	if (task == NULL)
 		return -ENODEV;
 
+	if (task->disconnecting == TRUE)
+		return -EINPROGRESS;
+
 	disable_network(task);
 
 	remove_network(task);
 
-	connman_network_set_connected(task->network, FALSE);
-
-	connman_network_unref(task->network);
+	task->disconnecting = TRUE;
 
 	return 0;
 }
