@@ -40,8 +40,11 @@ DBusMessage *__connman_error_not_implemented(DBusMessage *msg);
 DBusMessage *__connman_error_no_carrier(DBusMessage *msg);
 DBusMessage *__connman_error_in_progress(DBusMessage *msg);
 DBusMessage *__connman_error_already_connected(DBusMessage *msg);
+DBusMessage *__connman_error_not_connected(DBusMessage *msg);
 DBusMessage *__connman_error_operation_aborted(DBusMessage *msg);
 DBusMessage *__connman_error_operation_timeout(DBusMessage *msg);
+DBusMessage *__connman_error_invalid_service(DBusMessage *msg);
+DBusMessage *__connman_error_invalid_property(DBusMessage *msg);
 
 int __connman_selftest(void);
 
@@ -61,7 +64,7 @@ void __connman_profile_list(DBusMessageIter *iter);
 const char *__connman_profile_active_ident(void);
 const char *__connman_profile_active_path(void);
 
-void __connman_profile_changed(void);
+void __connman_profile_changed(gboolean delayed);
 
 #include <connman/log.h>
 
@@ -71,6 +74,8 @@ void __connman_log_cleanup(void);
 void __connman_toggle_debug(void);
 gboolean __connman_debug_enabled(void);
 
+#include <connman/option.h>
+
 #include <connman/plugin.h>
 
 int __connman_plugin_init(const char *pattern, const char *exclude);
@@ -81,13 +86,32 @@ void __connman_plugin_cleanup(void);
 int __connman_security_check_privilege(DBusMessage *message,
 				enum connman_security_privilege privilege);
 
-#include <connman/ipv4.h>
-
-const char *__connman_ipv4_method2string(enum connman_ipv4_method method);
-enum connman_ipv4_method __connman_ipv4_string2method(const char *method);
-
 #include <connman/ipconfig.h>
 
+const char *__connman_ipconfig_method2string(enum connman_ipconfig_method method);
+enum connman_ipconfig_method __connman_ipconfig_string2method(const char *method);
+
+void __connman_ipconfig_append_ipv4(struct connman_ipconfig *ipconfig,
+				DBusMessageIter *iter, const char *prefix);
+int __connman_ipconfig_set_ipv4(struct connman_ipconfig *ipconfig,
+				const char *key, DBusMessageIter *value);
+
+int __connman_ipconfig_load(struct connman_ipconfig *ipconfig,
+		GKeyFile *keyfile, const char *identifier, const char *prefix);
+int __connman_ipconfig_save(struct connman_ipconfig *ipconfig,
+		GKeyFile *keyfile, const char *identifier, const char *prefix);
+
+#include <connman/inet.h>
+
+enum connman_device_type __connman_inet_get_device_type(int index);
+
+#include <connman/wifi.h>
+
+#include <connman/rfkill.h>
+
+int __connman_rfkill_init(void);
+void __connman_rfkill_cleanup(void);
+
 #include <connman/resolver.h>
 
 int __connman_resolver_init(void);
@@ -141,6 +165,9 @@ struct connman_device *__connman_element_get_device(struct connman_element *elem
 const char *__connman_element_get_device_path(struct connman_element *element);
 const char *__connman_element_get_network_path(struct connman_element *element);
 
+struct connman_device *__connman_element_find_device(enum connman_device_type type);
+int __connman_element_request_scan(enum connman_device_type type);
+
 const char *__connman_element_type2string(enum connman_element_type type);
 
 static inline void __connman_element_lock(struct connman_element *element)
@@ -156,28 +183,18 @@ int __connman_element_append_ipv4(struct connman_element *element,
 int __connman_element_set_ipv4(struct connman_element *element,
 				const char *name, DBusMessageIter *value);
 
-int __connman_detect_init(void);
-void __connman_detect_cleanup(void);
-
 int __connman_ipv4_init(void);
 void __connman_ipv4_cleanup(void);
 
 int __connman_connection_init(void);
 void __connman_connection_cleanup(void);
 
-#ifdef HAVE_UDEV
+gboolean __connman_connection_update_gateway(void);
+
 int __connman_udev_init(void);
 void __connman_udev_cleanup(void);
-#else
-static inline int __connman_udev_init(void)
-{
-	return 0;
-}
-
-static inline void __connman_udev_cleanup(void)
-{
-}
-#endif
+char *__connman_udev_get_devtype(const char *ifname);
+char *__connman_udev_get_mbm_devnode(const char *ifname);
 
 #include <connman/device.h>
 
@@ -190,6 +207,7 @@ void __connman_device_decrease_connections(struct connman_device *device);
 void __connman_device_set_network(struct connman_device *device,
 					struct connman_network *network);
 
+int __connman_device_scan(struct connman_device *device);
 int __connman_device_connect(struct connman_device *device);
 int __connman_device_disconnect(struct connman_device *device);
 
@@ -211,6 +229,7 @@ void __connman_network_cleanup(void);
 void __connman_network_set_device(struct connman_network *network,
 					struct connman_device *device);
 
+int __connman_network_connect(struct connman_network *network);
 int __connman_network_disconnect(struct connman_network *network);
 
 connman_bool_t __connman_network_has_driver(struct connman_network *network);
@@ -237,16 +256,30 @@ struct connman_service *__connman_service_create_from_device(struct connman_devi
 struct connman_service *__connman_service_lookup_from_network(struct connman_network *network);
 struct connman_service *__connman_service_create_from_network(struct connman_network *network);
 
+unsigned int __connman_service_get_order(struct connman_service *service);
+
 int __connman_service_set_carrier(struct connman_service *service,
 						connman_bool_t carrier);
 int __connman_service_indicate_state(struct connman_service *service,
 					enum connman_service_state state);
+int __connman_service_indicate_error(struct connman_service *service,
+					enum connman_service_error error);
+int __connman_service_indicate_default(struct connman_service *service);
+
+int __connman_service_connect(struct connman_service *service);
+int __connman_service_disconnect(struct connman_service *service);
+int __connman_service_create_and_connect(DBusMessage *msg);
+void __connman_service_auto_connect(void);
 
 #include <connman/notifier.h>
 
 int __connman_notifier_init(void);
 void __connman_notifier_cleanup(void);
 
+void __connman_notifier_device_type_list(gboolean powered,
+						DBusMessageIter *iter);
+void __connman_notifier_device_type_register(enum connman_device_type type);
+void __connman_notifier_device_type_unregister(enum connman_device_type type);
 void __connman_notifier_device_type_increase(enum connman_device_type type);
 void __connman_notifier_device_type_decrease(enum connman_device_type type);
 void __connman_notifier_offline_mode(connman_bool_t enabled);
