@@ -330,6 +330,89 @@ const char *__connman_element_get_network_path(struct connman_element *element)
 	return __connman_element_get_network_path(element->parent);
 }
 
+struct find_data {
+	enum connman_device_type type;
+	struct connman_device *device;
+};
+
+static gboolean find_device(GNode *node, gpointer user_data)
+{
+	struct connman_element *element = node->data;
+	struct find_data *data = user_data;
+
+	if (element->type != CONNMAN_ELEMENT_TYPE_DEVICE)
+		return FALSE;
+
+	if (element->device == NULL)
+		return FALSE;
+
+	if (data->type != connman_device_get_type(element->device))
+		return FALSE;
+
+	data->device = element->device;
+
+	return TRUE;
+}
+
+struct connman_device *__connman_element_find_device(enum connman_device_type type)
+{
+	struct find_data data = { .type = type, .device = NULL };
+
+	g_node_traverse(element_root, G_PRE_ORDER,
+				G_TRAVERSE_ALL, -1, find_device, &data);
+
+	return data.device;
+}
+
+static gboolean request_scan(GNode *node, gpointer user_data)
+{
+	struct connman_element *element = node->data;
+	struct find_data *data = user_data;
+	enum connman_device_type type;
+
+	if (element->type != CONNMAN_ELEMENT_TYPE_DEVICE)
+		return FALSE;
+
+	if (element->device == NULL)
+		return FALSE;
+
+	type = connman_device_get_type(element->device);
+
+	switch (type) {
+	case CONNMAN_DEVICE_TYPE_UNKNOWN:
+	case CONNMAN_DEVICE_TYPE_VENDOR:
+	case CONNMAN_DEVICE_TYPE_ETHERNET:
+	case CONNMAN_DEVICE_TYPE_BLUETOOTH:
+	case CONNMAN_DEVICE_TYPE_GPS:
+	case CONNMAN_DEVICE_TYPE_MBM:
+	case CONNMAN_DEVICE_TYPE_HSO:
+	case CONNMAN_DEVICE_TYPE_NOZOMI:
+	case CONNMAN_DEVICE_TYPE_HUAWEI:
+	case CONNMAN_DEVICE_TYPE_NOVATEL:
+		return FALSE;
+	case CONNMAN_DEVICE_TYPE_WIFI:
+	case CONNMAN_DEVICE_TYPE_WIMAX:
+		if (data->type != CONNMAN_DEVICE_TYPE_UNKNOWN &&
+							data->type != type)
+			return FALSE;
+		break;
+	}
+
+	__connman_device_scan(element->device);
+
+	return FALSE;
+}
+
+int __connman_element_request_scan(enum connman_device_type type)
+{
+	struct find_data data = { .type = type, .device = NULL };
+
+	g_node_traverse(element_root, G_PRE_ORDER,
+				G_TRAVERSE_ALL, -1, request_scan, &data);
+
+	return 0;
+}
+
 static gint compare_priority(gconstpointer a, gconstpointer b)
 {
 	const struct connman_driver *driver1 = a;
@@ -774,7 +857,7 @@ static gboolean get_static_array_property(struct connman_element *element,
 
 	property = g_hash_table_lookup(element->properties, name);
 	if (property != NULL) {
-		*((char **) value) = property->value;
+		*((void **) value) = property->value;
 		*len = property->size;
 		found = TRUE;
 	}
@@ -1336,6 +1419,8 @@ static enum connman_service_error convert_error(enum connman_element_error error
 		break;
 	case CONNMAN_ELEMENT_ERROR_DHCP_FAILED:
 		return CONNMAN_SERVICE_ERROR_DHCP_FAILED;
+	case CONNMAN_ELEMENT_ERROR_CONNECT_FAILED:
+		return CONNMAN_SERVICE_ERROR_CONNECT_FAILED;
 	}
 
 	return CONNMAN_SERVICE_ERROR_UNKNOWN;
